require 'concurrent-ruby'
require 'time'
require "schdlr/version"
require "schdlr/dispatcher"
require "schdlr/job"
require "schdlr/queue/sqlite"

class Schdlr
  class Error < StandardError; end

  class JobFailed < Error
    attr_accessor :task, :reason
    def initialize(task, reason=nil)
      @task = task
      @reason = reason
    end
  end

  def initialize(queue, **options)
    @queue = queue
    @dispatcher = Dispatcher.new
    execution_interval = options[:interval] || 1
    timeout_interval = options[:timeout] || 1
    @timer = Concurrent::TimerTask.new(execution_interval: execution_interval, timeout_interval: timeout_interval) { puts Time.now }
    @timer.add_observer(self)
  end

  def start!
    @timer.execute
  end

  def stop!
    @timer.shutdown
  end

  def at(datetime, task)
    @queue.set(datetime, task)
  end

  def now(task)
    at(Time.now, task)
  end

  def task_update(time, result, ex)
    if ex
      id, _, _, _ = *ex.task
      @queue.fail(id)
    else
      id, _, _, _ = *result
      @queue.del(id)
    end
  end

  def update(time, result, ex)
    tasks = @queue.tasks(time)
    tasks.collect do |task_spec|
      ivar = @dispatcher.async.dispatch(task_spec)
      ivar.add_observer(self, :task_update)
      ivar
    end
  end
end
